#!/usr/bin/env python
'''Module to test the significance of a correlation coeffictent difference

Consider three samples of the same length: Y, X1, X2
then 

    correlationcoefficient_1 := rho(Y, X1)
    correlationcoefficient_2 := rho(Y, X2)

with rho the Pearson correlation coeffictent.

This module can be used to test whether the difference between two dependent
correlation coeffients is significant.

To this end, this module produces the probabilities for the tests:

H0 : correlationcoefficient_1 <= correlationcoefficient_2 (one tailed)

and

H0 : correlationcoefficient_1 == correlationcoefficient_2 (two tailed)


Depends on Numpy, Scipy (matplotlib optional)


r significance testing based on:
Kendall, M., & Stuart, A. (1979). The Advanced Theory of Statistics, vol. 2: Inference and Relationship. Charles Griffin & Company Ltd., London, fourth ed.
which is probably based on Fisher, R. A. (1925). Applications of "Student's" distribution. Metron, 5 , 90-104.


r difference testing based on:
Meng, X.-L., Rosenthal, R., & Rubin, D. B. (1992). Comparing correlated correlation coefficients. Psychological Bulletin, 111(1), 172-175.


DEPENDENCIES
    Numpy
    Scipy
    matplotlib (optional)


Copyright (c) 2012 CLiPS. All rights reserved.

# License: GNU General Public License, see http://www.clips.ua.ac.be/~vincent/scripts/LICENSE.txt
'''
__version__='1.1.1'
__author__='Vincent Van Asch'
__date__ = 'March 2012'

import os, sys, getopt
from math import pow, sqrt

from numpy import average, arctanh
from scipy.stats import norm
from scipy.stats import t as student

try:
    import pylab
except ImportError:
    print >>sys.stderr, 'The plotting option is not available because matplotlib is not installed.'
    PLOT=False
else:
    PLOT=True


class Data(list):
    '''A data list'''
    def __init__(self, name, d=[]):
        list.__init__(self, d)
        self._name = name
    @property
    def name(self):
        '''The name of the data'''
        return self._name


def fread(fname):
    with open(os.path.abspath(os.path.expanduser(fname)), 'rU') as f:
        for l in f:
            line = l.strip()
            if line and line[0] != '#': 
                yield line
                
                    
def correlation(x,y):
    '''Takes two paired lists and returns the 

    Pearson product-moment correlation coefficient
    
    from page 61:
    
    Thirteen Ways to Look at the Correlation Coefficient
    Joseph Lee Rodgers and W. Alan Nicewander
    The American Statistician
    Vol. 42, No. 1 (Feb., 1988), pp. 59-66
    Published by: American Statistical Association
    Article Stable URL: http://www.jstor.org/stable/2685263
    '''
    if (not x) or (not y): raise ValueError('x nor y can be empty.')
    if len(x) != len(y): raise ValueError('x (%d) and y (%d) should have the same length.' %(len(x), len(y)))
    
    xmean = average(x)
    ymean = average(y)
    
    n=len(x)
    
    X=0
    Y=0
    T=0
    for i in range(n):
        xdiff = x[i]-xmean
        ydiff = y[i]-ymean
        
        xdiff2 = pow(xdiff, 2)
        ydiff2 = pow(ydiff, 2)
    
        T += (xdiff*ydiff)
        X += xdiff2
        Y += ydiff2
        
    N = sqrt(X*Y)
        
    try:
        return T/N
    except ZeroDivisionError:
        print 'x', x
        print 'y', y
        raise
        
        
        
def significance(r, n, verbose=False):
    '''
    Takes a correlation coefficent r and the
    number of observations n and returns
    the probability for the test
    
    H0 : r == 0
    '''
    if n < 6:
        print >>sys.stderr, 'WARNING: for a good result the number of observations should be >= 6!'
    
    # The test statistic
    df = n-2.0
    t = r / sqrt((1.0-pow(r, 2))/df)
    if verbose: print >>sys.stderr, 't-score:', t
        
    # The probability
    p = 2.0*(1.0-student.cdf(abs(t), df))
    
    return p
        
        
        
def probability(r1, r2, r, n, two_tailed=False, verbose=False):
    '''
    Takes for three samples Y, X1, X2
    r1: correlation coefficient rho(Y, X1)
    r2: correlation coefficient rho(Y, X2)
    r : correlation coefficient rho(X1, X2)
    n : the size of the samples
    
    Returns the one-sided probability for 
    
    H0 : r1 - r2 = diff <= 0 (i.e. r1 <= r2)
    
    This means that of the returned probability is lower
    than typically 0.05 that we reject H0 and
    say that significantly r1 > r2.
    
    if two_tailed == True:
    
    For the two-sided test
    H0 : r1 - r2 = diff == 0 (i.e. r1 == r2)
    
    the returned probability can be compared
    to a predefined signifance level.
    
    verbose: print the Z-score
    '''
    # The Fisher z transformation
    z1 = arctanh(r1)
    z2 = arctanh(r2)
    
    # The mean r2
    mr2 = 0.5*(pow(r1, 2) + pow(r2, 2))
    
    # The f factor
    f = (1.0-r)/(2.0*(1.0-mr2))
    if f > 1: f = 1.0
    
    # The h factor
    h = (1.0-f*mr2)/(1.0-mr2)
    
    # The Z score
    Z = (z1 - z2)*sqrt( (n-3.0) / (2.0*(1.0 - r)*h) )
    if verbose: print >>sys.stderr, 'Z-score:', Z
    
    # The one-tailed probability
    p = norm.cdf(Z)
    
    if not two_tailed:
        return 1.0-p
    else:
        return (1-norm.cdf(abs(Z)))*2.0

    
def readfile(fname):
    '''Takes a file with on each line
    a float'''
    data = Data(fname)
    for line in fread(fname):
        data.append(float(line))
    return data
    

def test(y, x1, x2, two_tailed=False, verbose=False, plot=False):
    '''Takes 3 Data objects and returns the probability()'''
    r1 = correlation(y, x1)
    r2 = correlation(y, x2)
    r = correlation(x1, x2)
    
    if verbose:
        print 'Number of observations : %d' %(len(y))
        print 'Correlation r1 : %.5f (%s -- %s)' %(r1, os.path.basename(y.name), os.path.basename(x1.name))
        print 'Correlation r2 : %.5f (%s -- %s)' %(r2, os.path.basename(y.name), os.path.basename(x2.name))
        print 'Correlation r  : %.5f (%s -- %s)' %(r , os.path.basename(x1.name), os.path.basename(x2.name))
    
    n = len(y)
    p  = probability(r1, r2, r, n, two_tailed=two_tailed, verbose=verbose)
    
    if plot and PLOT:
        pylab.figure()
        pylab.plot(x1, y, 'r+')
        pylab.plot(x2, y, 'g+')
        pylab.xlabel('X1 (red) and X2 (green)')
        pylab.ylabel('Y')
        pylab.title('%d observations' %n)
    
    return p
    


def filesignificance(yname, xname, verbose=False, plot=False):
    '''
    Takes a two filenames of equal length
    and returns the significance()
    '''
    # Read in the files
    y = readfile(yname)
    x = readfile(xname)
    
    # The correlation
    n = len(y)
    r = correlation(y, x)
    
    if verbose:
        print 'Number of observations : %d' %(n)
        print 'Correlation r : %.5f (%s -- %s)' %(r, os.path.basename(y.name), os.path.basename(x.name))
    
    if plot and PLOT:
        pylab.figure()
        pylab.plot(x, y, 'r+')
        pylab.xlabel('X')
        pylab.ylabel('Y')
        pylab.title('%d observations' %n)
        
    # The probability
    return significance(r, n, verbose=verbose)

def filetest(yname, xname1, xname2, two_tailed=False, verbose=False, plot=False):
        '''
        Takes 3 filenames for
        files of equal length.
        and returns the probability()
        '''
        # Read in files
        y = readfile(yname)
        x1 = readfile(xname1)
        x2 = readfile(xname2)
        return test(y, x1, x2, two_tailed=two_tailed, verbose=verbose, plot=plot)
        
        
        
def main1(yname, xname, verbose=False, plot=False):
    '''Main1'''
    p = filesignificance(yname, xname, verbose=verbose, plot=plot)
    
    if verbose:
        print '''\nH0 : r  == 0
Two-tailed probability: %f''' %(p)
        print '''\nNormally, you would reject H0 if the probabilty is below a predefined
significance level, say 0.05.'''
    else:
        print p
    
    if plot and PLOT:
        pylab.show()
    
def main2(r, n, verbose=False):
    '''Main2'''
    p = significance(r, n, verbose=False)
    
    if verbose:
        print '''\nH0 : r  == 0
Two-tailed probability: %f''' %(p)
        print '''\nNormally, you would reject H0 if the probabilty is below a predefined
significance level, say 0.05.'''
    else:
        print p
        
        
def main3(yname, xname1, xname2, two_tailed=False, verbose=False, plot=False):
    '''Main'''
    # Get the probality
    p = filetest(yname, xname1, xname2, two_tailed=two_tailed, verbose=verbose, plot=plot)
    
    # Report
    if verbose:
        print
        if two_tailed:
            print '''H0 : r1 - r2 == 0
Two-tailed probability: %f''' %(p)
        else:
            print '''H0: r1 -r2 <= 0
One-tailed probability: %f''' %(p)
            
        print '''\nNormally, you would reject H0 if the probabilty is below a predefined
significance level, say 0.05.'''
    else:
        print p
        
    if plot and PLOT:
        pylab.show()
        
        
def main4(r1, r2, r, n, two_tailed=False, verbose=False):
    '''Main4'''
    # The probability
    p =  probability(r1, r2, r, n, two_tailed=two_tailed, verbose=verbose)
        
    # Report
    if verbose:
        print
        if two_tailed:
            print '''H0 : r1 - r2 == 0
Two-tailed probability: %f''' %(p)
        else:
            print '''H0: r1 -r2 <= 0
One-tailed probability: %f''' %(p)
            
        print '''\nNormally, you would reject H0 if the probabilty is below a predefined
significance level, say 0.05.'''
    else:
        print p
        
        
        
def _usage():
    print >>sys.stderr, '''Correlation coefficient significance testing (version %s)
    
USAGE
    python rtest.py [-p] [-v] [-t] [-f] args
        
        
    The script can be used to test the significance of a correlation coefficient
    and to test the difference between two correlation coefficients.
        
    The functionality depends on the number of arguments you supply:
    
    * 2 arguments and option -f: args = y, x
    
        y, x:
        Two pathnames to files of equal length.
        Each line in the files should contain one float.
        
        Computes the correlation coefficient r and prints
        the probability for the test: H0 : r == 0
        
    * 2 arguments without option -f: args = r, n
    
        r : corr coef between two variables
        n : number of observations 
        
        Prints the probability for the test:
        H0 : r == 0
        
    * 3 arguments: args = y, x1, x2
    
        y, x1, x2:
        Three pathnames to files of equal length.
        Each line in the files should contain one float.
        
        Computes the required correlation coefficients
        and prints the probability.  
    
        Depending on option -t the test is either:
            H0: r1 -r2 <= 0
        or
            H0 : r1 - r2 == 0
        
    * 4 arguments: args = r1 r2 rx n
    
        r1: corr coef between y and x1
        r2: corr coef between y and x2
        rx: corr coef between x1 and x2
        n : number of observations
        
        Prints the probability.
        
        Depending on option -t the test is either:
            H0: r1 -r2 <= 0
        or
            H0 : r1 - r2 == 0
        
        
OPTIONS
    -t : Report two-tailed probality
    -v : Print more information
    -p : Plot the data points. Only available when 1 or 4 arguments
         are given (matplotlib needed)
    
%s, %s''' %(__version__, __author__, __date__)
    
        
if __name__ == '__main__':
    try:
        opts,args=getopt.getopt(sys.argv[1:],'hftvp', ['help'])
    except getopt.GetoptError:
        # print help information and exit:
        _usage()
        sys.exit(2)

    two_tailed=False
    verbose=False
    plot=False
    
    files = False
    
    for o, a in opts:
        if o in ('-h', '--help'):
            _usage()
            sys.exit()
        if o == '-t':
            two_tailed=True
        if o == '-v':
            verbose=True
        if o == '-p':
            plot=True
        if o == '-f':
            files=True


    mode = len(args)

    if mode == 2 and files:
        y = os.path.abspath(os.path.expanduser(args[0]))
        x = os.path.abspath(os.path.expanduser(args[1]))
        main1(y, x, verbose=verbose, plot=plot)
    elif mode == 2:
        r = float(args[0])
        n = int(args[1])
        main2(r, n, verbose=verbose)
    elif mode == 3:
        y, x1, x2 = args
        main3(y, x1, x2, two_tailed=two_tailed, verbose=verbose, plot=plot)
    elif mode == 4:
        r1, r2, r, n = [float(x) for x in args]
        main4(r1, r2, r, n, two_tailed=two_tailed, verbose=verbose)
    else:
        _usage()
        sys.exit(1)
        
    
        
    
        
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    

    
        